from abc import abstractmethod
from typing import AsyncIterator, Optional

from langchain_core.messages import AIMessageChunk
from langchain_core.runnables import RunnableConfig
from langchain_core.tracers.context import collect_runs
from langgraph.prebuilt.tool_executor import ToolExecutor, ToolInvocation

from docugami_langchain.agents.models import (
    AgentState,
    Citation,
    CitedAnswer,
    Invocation,
    StepState,
)
from docugami_langchain.base_runnable import BaseRunnable, TracedResponse
from docugami_langchain.tools.common import BaseDocugamiTool

THINKING = "Thinking..."


class BaseDocugamiAgent(BaseRunnable[AgentState]):
    """
    Base class with common functionality for various chains.
    """

    tools: list[BaseDocugamiTool] = []

    @abstractmethod
    def streamable_node_names(self) -> list[str]:
        """Node names in the graph from which token by token output should be streamed."""
        ...

    @abstractmethod
    def parse_final_answer_from_streamed_output(self, text: str) -> str:
        """Given output stream from a streamable node, parses out the final answer (e.g. past a delimiter)."""
        ...

    def execute_tool(
        self,
        state: AgentState,
        config: Optional[RunnableConfig],
    ) -> AgentState:
        """
        Gets the most recent tool invocation (added by the agent) and execute it.
        """

        inv_model = state.get("tool_invocation")
        if not inv_model:
            raise Exception(f"No tool invocation in model: {state}")

        previous_steps = state.get("intermediate_steps") or []
        if previous_steps and any(
            [s for s in previous_steps if s.invocation == inv_model]
        ):
            tool_output = CitedAnswer(
                source=inv_model.tool_name,
                answer="This tool has been invoked before with identical inputs. Please try different inputs or a different tool, after reconsidering previous thoughts and observations. "
                + "Be careful you don't get stuck in a loop.",
            )
        else:
            tool_executor = ToolExecutor(self.tools)
            tool_output = tool_executor.invoke(
                ToolInvocation(  # LangChain version of the Invocation object
                    tool=inv_model.tool_name,
                    tool_input=inv_model.tool_input,
                ),
                config,
            )
            if not isinstance(tool_output, CitedAnswer):
                raise Exception(f"Invalid output from tool executor: {tool_output}")

        step = StepState(
            invocation=inv_model,
            output=tool_output.answer,
            citations=tool_output.citations,
        )
        return {"intermediate_steps": previous_steps + [step]}

    def invocation_answer(
        self,
        invocation: Invocation,
        answer_source: str,
    ) -> AgentState:
        """
        Builds a human readable interim answer from a tool invocation.
        """

        tool_name = invocation.tool_name
        tool_input = invocation.tool_input
        if tool_name and tool_input:
            busy_text = THINKING
            match = [t for t in self.tools if t.name.lower() == tool_name]
            if match:
                busy_text = match[0].to_human_readable(invocation)

        return {
            "tool_invocation": invocation,
            "cited_answer": CitedAnswer(
                source=answer_source,
                answer=busy_text,  # Show the user interim output.
            ),
        }

    def run(  # type: ignore[override]
        self,
        question: str,
        chat_history: list[tuple[str, str]] = [],
        config: Optional[RunnableConfig] = None,
    ) -> TracedResponse[AgentState]:
        if not question:
            raise Exception("Input required: question")

        return super().run(
            question=question,
            chat_history=chat_history,
            config=config,
        )

    def run_batch(  # type: ignore[override]
        self,
        inputs: list[tuple[str, list[tuple[str, str]]]],
        config: Optional[RunnableConfig] = None,
    ) -> list[AgentState]:
        return super().run_batch(
            inputs=[
                {
                    "question": i[0],
                    "chat_history": i[1],
                }
                for i in inputs
            ],
            config=config,
        )

    async def run_stream(  # type: ignore[override]
        self,
        question: str,
        chat_history: list[tuple[str, str]] = [],
        config: Optional[RunnableConfig] = None,
    ) -> AsyncIterator[TracedResponse[AgentState]]:
        if not question:
            raise Exception("Input required: question")

        config, kwargs_dict = self._prepare_run_args(
            {
                "question": question,
                "chat_history": chat_history,
            }
        )

        with collect_runs() as cb:
            last_response_value = None
            current_step_token_stream = ""
            final_streaming_started = False
            citations: list[Citation] = []
            async for event in self.runnable().astream_events(
                input=kwargs_dict, config=config, version="v1"
            ):
                event_key = event.get("event")
                event_name = event.get("name")
                event_data = event.get("data")
                if event_data:
                    if event_name in self.streamable_node_names():
                        if event_key == "on_chain_start":
                            # Restart token stream every time a streamable node starts
                            current_step_token_stream = ""
                            final_streaming_started = True
                        elif event_key == "on_chain_end":
                            # Yield the completed output when a streamable node finishes
                            last_response_value = event_data.get("output")
                            if last_response_value:
                                yield TracedResponse[AgentState](
                                    value=last_response_value
                                )
                    elif event_name == "execute_tool":
                        if event_key == "on_chain_end":
                            state: Optional[AgentState] = event_data.get("output")
                            if state:
                                answer = state.get("cited_answer")
                                citations = answer.citations if answer else []
                    elif event_key == "on_chat_model_stream":
                        chunk = event_data.get("chunk")
                        if isinstance(chunk, AIMessageChunk):
                            current_step_token_stream += str(chunk.content)
                            final_answer = self.parse_final_answer_from_streamed_output(
                                current_step_token_stream
                            )

                            if final_answer:
                                # Source the answer from the last step, if any
                                if not final_streaming_started:
                                    # Set final streaming started once as soon as we see the final answer
                                    final_streaming_started = bool(final_answer)
                                else:
                                    # Start streaming the final answer, no more interim steps
                                    last_response_value = AgentState(
                                        chat_history=[],
                                        question="",
                                        tool_invocation=None,
                                        cited_answer=CitedAnswer(
                                            source=self.__class__.__name__,
                                            answer=final_answer,
                                            citations=citations,
                                            is_final=True,
                                        ),
                                    )
                                    yield TracedResponse[AgentState](
                                        value=last_response_value
                                    )

            # Yield the final result with the run_id
            if last_response_value:
                if "cited_answer" in last_response_value:
                    last_response_value["cited_answer"].is_final = True

            if cb.traced_runs:
                run_id = str(cb.traced_runs[0].id)
                yield TracedResponse[AgentState](
                    run_id=run_id,
                    value=last_response_value,  # type: ignore
                )



from docugami_langchain.chains.rag.simple_rag_chain import SimpleRAGChain
from docugami_langchain.chains.rag.standalone_question_chain import (
    StandaloneQuestionChain,
)
from docugami_langchain.chains.rag.suggested_questions_chain import (
    SuggestedQuestionsChain,
)
from docugami_langchain.chains.rag.suggested_report_chain import SuggestedReportChain
from docugami_langchain.chains.rag.tool_final_answer_chain import ToolFinalAnswerChain
from docugami_langchain.chains.rag.tool_output_grader_chain import ToolOutputGraderChain

__all__ = [
    "SimpleRAGChain",
    "StandaloneQuestionChain",
    "SuggestedQuestionsChain",
    "SuggestedReportChain",
    "ToolFinalAnswerChain",
    "ToolOutputGraderChain",
]



from dataclasses import dataclass, field
from typing import Optional

from langchain_core.runnables import Runnable

from docugami_langchain.config import DEFAULT_EXAMPLES_PER_PROMPT


@dataclass
class RunnableSingleParameter:
    variable: str
    key: str
    description: str


@dataclass
class RunnableParameters:
    inputs: list[RunnableSingleParameter]
    output: RunnableSingleParameter
    task_description: str
    additional_instructions: list[str] = field(default_factory=lambda: [])
    stop_sequences: list[str] = field(default_factory=lambda: ["\n", "<|eot_id|>"])
    num_examples: int = DEFAULT_EXAMPLES_PER_PROMPT
    additional_runnables: Optional[list[Runnable]] = None
    key_finding_output_parse: bool = True
    include_output_instruction_suffix: bool = False


__all__ = ["RunnableSingleParameter", "RunnableParameters"]



from typing import Optional

from langchain_core.output_parsers import PydanticOutputParser
from langchain_core.runnables import Runnable, RunnableConfig
from langgraph.graph import END, StateGraph

from docugami_langchain.agents.base import BaseDocugamiAgent
from docugami_langchain.agents.models import (
    AgentState,
    CitedAnswer,
    Invocation,
)
from docugami_langchain.chains.rag import (
    StandaloneQuestionChain,
    ToolFinalAnswerChain,
    ToolOutputGraderChain,
)
from docugami_langchain.history import steps_to_str
from docugami_langchain.output_parsers import TextCleaningOutputParser
from docugami_langchain.params import RunnableParameters, RunnableSingleParameter
from docugami_langchain.tools.common import render_text_description


class ToolRouterAgent(BaseDocugamiAgent):
    """
    Agent that implements agentic RAG with a tool router implementation.
    """

    standalone_question_chain: StandaloneQuestionChain
    output_grader_chain: ToolOutputGraderChain
    final_answer_chain: ToolFinalAnswerChain

    def params(self) -> RunnableParameters:
        """The params are directly implemented in the runnable."""
        return RunnableParameters(
            inputs=[
                RunnableSingleParameter(
                    "question",
                    "QUESTION",
                    "Question asked by the user, which must be answered from one of the given tools.",
                ),
                RunnableSingleParameter(
                    "tool_names",
                    "TOOL NAMES",
                    "List (names) of tools that you must exclusively pick one from, in order to answer the given question.",
                ),
                RunnableSingleParameter(
                    "tool_descriptions",
                    "TOOL DESCRIPTIONS",
                    "Detailed description of tools that you must exclusively pick one from, in order to answer the given question.",
                ),
                RunnableSingleParameter(
                    "intermediate_steps",
                    "INTERMEDIATE STEPS",
                    "The inputs and outputs to various intermediate steps an AI agent has previously taken to consider the question using specialized tools. "
                    + "Carefully consider these intermediate steps as you decide on your next tool invocation, think step by step!",
                ),
            ],
            output=RunnableSingleParameter(
                "tool_invocation_json",
                "TOOL INVOCATION JSON",
                "A JSON blob with the name of the tool to use (`tool_name`) and the input to send it per the tool description (`tool_input`)",
            ),
            task_description="selects an appropriate tool for the question a user is asking, and builds a tool invocation JSON blob for the tool",
            additional_instructions=[
                """- Here is an example of a valid JSON blob for your output. Please STRICTLY follow this format:
{{
  "tool_name": $TOOL_NAME,
  "tool_input": $INPUT_STRING
}}""",
                "- Always use one of the tools, don't try to directly answer the question even if you think you know the answer",
                "- $TOOL_NAME is the (string) name of the tool to use, and must be one of these values: {tool_names}",
                "- $INPUT_STRING is the (string) input carefully crafted to answer the question using the given tool.",
                "- Before retrying a tool, look at previous attempts at running the tool (in intermediate steps) and try to update the inputs to the tool before trying again",
            ],
            stop_sequences=["<|eot_id|>"],
            additional_runnables=[TextCleaningOutputParser(), PydanticOutputParser(pydantic_object=Invocation)],  # type: ignore
        )

    def runnable(self) -> Runnable:
        """
        Custom runnable for this agent.
        """

        def run_agent(
            state: AgentState, config: Optional[RunnableConfig]
        ) -> AgentState:
            return {
                "tool_names": ", ".join([t.name for t in self.tools]),
                "tool_descriptions": "\n" + render_text_description(self.tools),
                "intermediate_steps": [],
            }

        tool_invocation_runnable: Runnable = {
            "question": lambda x: x["question"],
            "tool_names": lambda x: x["tool_names"],
            "tool_descriptions": lambda x: x["tool_descriptions"],
            "intermediate_steps": lambda x: steps_to_str(x["intermediate_steps"]),
        } | super().runnable()

        def standalone_question(
            state: AgentState, config: Optional[RunnableConfig]
        ) -> AgentState:
            question = state.get("question")
            chat_history = state.get("chat_history")

            if question and chat_history:
                standalone_question_response = self.standalone_question_chain.run(
                    question, chat_history, config
                )

                if standalone_question_response.value:
                    state["question"] = standalone_question_response.value

            return state

        def generate_tool_invocation(
            state: AgentState, config: Optional[RunnableConfig]
        ) -> AgentState:
            invocation: Invocation = tool_invocation_runnable.invoke(state, config)
            answer_source = ToolRouterAgent.__name__

            return self.invocation_answer(invocation, answer_source)

        def grade_output(state: AgentState, config: Optional[RunnableConfig]) -> str:
            question = state.get("question")
            tool_descriptions = state.get("tool_descriptions")
            intermediate_steps = state.get("intermediate_steps")

            if question and tool_descriptions:
                if not intermediate_steps:
                    intermediate_steps = []

                grader_output_response = self.output_grader_chain.run(
                    question, tool_descriptions, intermediate_steps, config
                )

                grader_output_result = grader_output_response.value.lower()
                if "true" in grader_output_result or "yes" in grader_output_result:
                    return "true"

            return "false"

        def generate_final_answer(
            state: AgentState, config: Optional[RunnableConfig]
        ) -> AgentState:
            question = state.get("question")
            tool_descriptions = state.get("tool_descriptions")
            intermediate_steps = state.get("intermediate_steps")

            if not question or not tool_descriptions or not intermediate_steps:
                raise Exception("incomplete inputs")

            final_answer_response = self.final_answer_chain.run(
                question, tool_descriptions, intermediate_steps, config
            )

            # Source the answer from the last step, if any
            answer_source = ToolRouterAgent.__name__
            citations = []
            if intermediate_steps:
                last_step = intermediate_steps[-1]
                answer_source = last_step.invocation.tool_name
                citations = last_step.citations
            return {
                "cited_answer": CitedAnswer(
                    source=answer_source,
                    is_final=True,
                    citations=citations,
                    answer=final_answer_response.value,
                ),
            }

        # Define a new graph
        workflow = StateGraph(AgentState)

        # Define the nodes of the graph
        workflow.add_node("run_agent", run_agent)  # type: ignore
        workflow.add_node("standalone_question", standalone_question)  # type: ignore
        workflow.add_node("generate_tool_invocation", generate_tool_invocation)  # type: ignore
        workflow.add_node("execute_tool", self.execute_tool)  # type: ignore
        workflow.add_node("generate_final_answer", generate_final_answer)  # type: ignore

        # Set the entrypoint
        workflow.set_entry_point("run_agent")

        # Add edges
        workflow.add_edge("run_agent", "standalone_question")
        workflow.add_edge("standalone_question", "generate_tool_invocation")
        workflow.add_edge("generate_tool_invocation", "execute_tool")
        workflow.add_edge("generate_final_answer", END)

        # Decide whether to keep iterating or generate final answer
        workflow.add_conditional_edges(
            "execute_tool",
            grade_output,
            {
                "true": "generate_final_answer",
                "false": "generate_tool_invocation",  # try again
            },
        )

        # Compile
        return workflow.compile()

    def streamable_node_names(self) -> list[str]:
        """Node names in the graph from which token by token output should be streamed."""
        return ["generate_final_answer"]

    def parse_final_answer_from_streamed_output(self, text: str) -> str:
        """Given output stream from a streamable node, parses out the final answer (e.g. past a delimiter)."""
        return text  # no special delimiter in final answer
